
/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/*	
	computes the SKY value in an image file by inspecting the
	histogram near the max value

                SKY file [box=] [circle=] [bin=] [pix=] [quiet] [help]
	
	puts SKY and SKYSIG symbols into the PCVISTA table.

	1/30/91 - added 'floor' to "sky" expression, and <math.h>   - MWR

	2/21/92 - added exit(0) -rrt

	6/2/92  - Installed new method of calculating sky sigma -rrt 
	6/4/92  - allow negative values of sky -rrt
	9/22/92 - removed OLD_SKY trash -rrt
	10/22/92- removed small bug from 'fsky()' - we were including the
	             'topbin' value returned by the histogram routines as valid,
	             but it contains the number of all pixels with values
	             > OR = the top bin - thus, it can easily overwhelm the
	             true sky peak if there is a large tail of bright pixels. MWR
	10/29/92 - sky used without arguments gives usage -rrt
	12/15/92 - rewrote entirely to use gaussian fit to histogram in order
	             to determine the sky sigma; much faster than old method.
	             Also, with new, much larger NHIST, no need for 'bin'
	             parameter, so I've deleted it. MWR 
    12/17/92 - skysig is printed with fewer digits -rrt
	9/28/94  - circle added
			 - output now has keyword=value strings
			 - error messages try to contain program name -rrt
	1/8/96   - added 'help' option. MWR
	2/19/96  - now use fit to gaussian to determine sky value, as
	             well as sky-sigma value.  MWR
	7/13/96  - added "bin" option, allowing user to set bin size in
	             histogram.  MWR
*/

#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "pcvista.h"
#include "fits.h"

#ifndef DEBUG
#undef DEBUG					/* define this to get some diag output */
#endif

#ifdef PROTO
void main(int, char **);
static int find_sky(FITS_HANDLE fh, int sr, int sc, int nr, int nc, int nskip,
		    int binsize, int16 *sky, float *skysig);
static void make_hist(FITS_HANDLE fh, int sr, int sc, int nr, int nc, 
                      int nskip, int binsize, int min, int max);
static void find_range(int peak, double fraction, int *botbin, int *topbin);
static int find_fitted_values(int botbin, int topbin,
                              float *peak, float *sigma);
static float * vector(int nel);
#else
static int find_sky();
static void make_hist();
static void find_range();
static int find_fitted_values();
static float * vector();
#endif

#define BINSIZE   1        /* default binsize in sky histogram */
#define FRACTION     0.40  /* use only bins w/ at least this fraction */
#define MINBIN    5	   /* use at least this many bins around the peak */
                   	   /*   on either side of the peak, when fitting */
                   	   /*   a gaussian to find sky sigma             */

static long hist[NHIST + 1];


static int circle_flag;
static int cr,cc,rad;

#define USAGE "usage: sky file [box=] [circle=] [bin=] [pix=] [quiet] [help]"

void
main(argc, argv)
int argc;
char *argv[];
{
	int i, boxno=-1, nskip = 1, quiet_flag = 0, nrow, ncol,circno;
	int binsize = BINSIZE;
	int16 sky;
	FITS_HANDLE fh;
	static int nr, nc, sr, sc;
	char *gotit, buf[NBUF],fname[NBUF];
	float skysig;
	int gotfname=0;

	if (argc == 1) {
		error(-1, USAGE);
	}

	/* placate compiler */
	circno = 0;

	for (i = 1; i < argc; i++) {
		if (keyword("help", argv[i])) {
			fprintf(stdout,"%s\n",USAGE);
			exit(0);
		}
		if ((gotit = find("box", argv[i])) != NULL) {
			boxno = (int) (evaluate(gotit) + 0.5);
			continue;
		}
		if ((gotit = find("circle", argv[i])) != NULL) {
			circno = (int)(evaluate(gotit) + SMALL);
			if (getcircle(circno, &cr, &cc, &rad)){
				fprintf(stderr, "sky: circle %d not found\n",circno);
				exit(1);
			}
			circle_flag=1;
			continue;
		}
		if ((gotit = find("bin", argv[i])) != NULL) {
			binsize = (int) (evaluate(gotit) + 0.5);
			if (binsize < 1)
				error(-1, "bin parameter must be greater than or equal to 1");
			continue;
		}
		if ((gotit = find("pix", argv[i])) != NULL) {
			nskip = (int) (evaluate(gotit) + 0.5);
			if (nskip < 1)
				error(-1, "pix parameter must be greater than or equal to 1");
			continue;
		}
		if (keyword("quiet", argv[i])) {
			quiet_flag = 1;
			continue;
		}
		if(!gotfname){
			strcpy(fname,argv[i]);
			gotfname=1;
			continue;
		}
		fprintf(stderr,"sky: unknown argument '%s'\n",argv[i]);
		exit(1);
	}

	if(!gotfname)
		error(1,"no filename specified");

	fh = fits_open(fname, "r", &nrow, &ncol);
	nr = nrow;	
	nc = ncol;
	if(boxno !=-1){
		if(circle_flag)
			error(1,"sky: can't have circles and boxes");
		if (getbox(boxno, &sr, &sc, &nr, &nc))
			error(-1, "sky: box not found");
		check_box(boxno, nrow, ncol);
	}	
	if(circle_flag)
		check_circle(circno, nrow, ncol);
	
	if (find_sky(fh, sr, sc, nr, nc, nskip, binsize, &sky, &skysig) != 0) {
		error(0, "warning: sky or skysig may not be correct");
	}

	if (!quiet_flag) {
		printf("sky=%d skysig=%.2f\n", sky, skysig);
	}
	sprintf(buf, "sky=%d", sky);
	putsym(buf);
	sprintf(buf, "skysig=%.2f", skysig);
	putsym(buf);

	fits_close(fh);

	exit(0);
}



	/* given an image, find the a sky value for the image by looking at
	   the peak of the histogram of data values.  Use only those pixels 
	   in the general vicinity of the peak to find the variance of values
	   around the peak.

	   return 0 if all is well, or -1 if there's a problem.  
	   If all went well, place the sky value into the 'sky' parameter and
	   the skysigma into the 'skysig' parameter.  Otherwise, the values
	   of the two arguments are undefined (but probably 0.0). */

static int
find_sky(fh, sr, sc, nr, nc, nskip, binsize, sky, skysig)
FITS_HANDLE fh;
int sr, sc, nr, nc, nskip, binsize;
int16 *sky;
float *skysig;
{
	int i, peak, botbin, topbin, min, max, ret;
	long binmax;
	float skypeak;

	ret = 0;

	/* now construct the first histogram, using the min and max values 
	   as the first and last bins */
	min = MIN_DATA_VAL;
	max = MAX_DATA_VAL;

	make_hist(fh, sr, sc, nr, nc, nskip, binsize, min, max);

#ifdef DEBUG
	printf("  initial hist %-5d bins, min=%-7d  max=%-7d\n",
				NHIST, min, max);
#endif

	/* find the peak of the histogram, and set the sky equal to the value
	   at the peak. However, if the peak bin is the first, or last, issue
	   a warning error message, because the job was probably not done
	   correctly. */
	peak = 0;
	binmax = 0;
	for (i = 0; i < NHIST; i++) {
		if (hist[i] > binmax) {
			peak = i;
			binmax = hist[i];
		}
	}

        /*
         * before 2/16/96, we used to set the sky value equal to this
         * bin with the maximum number of pixels, like so:
                *sky = (int16) (min + peak);
         * but now we use the center of a fitted gaussian ... see below
         */

	if ((peak == 0) || (peak == (NHIST - 1))) {
		ret = -1;
	}	


	/* find the bins which contain at least FRACTION of the number of pixels
	   as the peak bin, centered on the peak of the histogram */
	find_range(peak, FRACTION, &botbin, &topbin);
#ifdef DEBUG
	printf("  botbin %-5d  peak %-5d  topbin %-5d\n", botbin, peak, topbin);
#endif

        /*
         * now we fit a gaussian to the pixel values in this range,
         * and use the fitted peak and sigma of the gaussian
         * for output we'll return to the user.
	 * 
	 * Don't forget to convert the fitted "skypeak" and "skysig"
	 * from _binned_ units back into real units!
         */
        if (find_fitted_values(botbin, topbin, &skypeak, skysig) != 0) {
                ret = -1;
        }
        *sky = (int16) (min + skypeak*binsize);
	*skysig *= binsize;

	return(ret);
}
	
	
	/* make a histogram for the given image, ignoring any points
	   smaller than 'min' or larger than 'max' */

static void
make_hist(fh, sr, sc, nr, nc, nskip, binsize, min, max)
FITS_HANDLE fh;
int sr, sc, nr, nc, nskip, binsize, min, max;
{
	int i, j, bin;
	int16 val, data[NMAX + 1];

	for (i = 0; i < NHIST; i++)
		hist[i] = 0;

	if(circle_flag){
		sr=cr-rad;
		nr=2*rad;
	}

	for (i = 0; i < nr; i++) {
		if(circle_flag){
			int chord;
			chord=find_chord(sr+i-cr,rad);
			nc=2*chord;
			sc=cc-chord;
		}
		fits_get_data(fh, sr + i, sc, data, nc);
		for (j = 0; j < nc; j += nskip) {
			val = data[j];
			if ((val >= min) && (val <= max)) {
				bin = (int) ((val - min)/binsize);
				hist[bin]++;
			}
		}
	}
}


	/* by going through the histogram array (assumed to have been
	   calculated already!), find the range of bin values, centered 
	   on the 'peak'th bin, between which the bins have at least FRACTION
	   times the number of counts of the peak bin. return the min and max 
	   bin in the 'botbin' and 'topbin' parameters. */


static void
find_range(peak, fraction, botbin, topbin)
int peak, *botbin, *topbin;
double fraction;
{
	int top, bot, cut;

	cut = FRACTION*hist[peak];

	for (bot = peak - MINBIN; bot > 0; bot--)
		if (hist[bot] < cut)
			break;
	for (top = peak + MINBIN; top < (NHIST - 1); top++)
		if (hist[top] < cut)
			break;

	*botbin = bot;
	*topbin = top;
}



        /* find the stdev of all pixels which fall into the bins between
         * 'botbin' and 'topbin', inclusive, and return the number.
         * to find the stdev, fit a gaussian to the histogram values
         * in the given range and return the gaussian's "peak" and "sigma".
         *
         * Return
         *     0  if all goes well, or
         *    -1  if the standard deviation is undefined, or if we
         *             can't allocate for memory.
         */

static int
find_fitted_values(botbin, topbin, peak, sigma)
int topbin, botbin;
float *peak, *sigma;
{
	int i, ndata;
	float *x, *y, *sig, amp, center, width, chisq;

	if ((ndata = (topbin - botbin) + 1) < 3) {
		error(0, "find_fitted_values: to few points to fit gaussian to sky");
		return(0.0);
	}	
	if ((x = vector(ndata)) == NULL) {
		error(0, "find_fitted_values: can't alloc for x");
		return(0.0);
	}
	if ((y = vector(ndata)) == NULL) {
		error(0, "find_fitted_values: can't alloc for y");
		free(x);
		return(0.0);
	}
	if ((sig = vector(ndata)) == NULL) {
		error(0, "find_fitted_values: can't alloc for sig");
		free(x);
		free(y);
		return(0.0);
	}
	for (i = 0; i < ndata; i++) {
		x[i] = i + botbin;
		y[i] = hist[i + botbin];
		sig[i] = 1.0;
	}
	if (gauss_fit(x, y, sig, ndata, &amp, &center, &width, &chisq) != 0) {
		error(0, "find_fitted_values: gaussFit returns with error");
		error(0, "find_fitted_values: setting sky sigma value to zero");
		width = 0.0;
		return(width);
	}

	free(x);
	free(y);
	free(sig);

	*peak = center;
	*sigma = width;

	return(0);
}
	

	/* return a pointer to an allocated array of floats, or NULL if 
	   we couldn't allocate memory */

static float *
vector(nel)
int nel;
{
	float *p;

	if ((p = (float *) malloc(nel*sizeof(float))) == NULL) {
		fprintf(stderr, "sky: can't allocate for array of %d floats", nel);
		return(NULL);
	}
	return(p);
}

